/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    Navigation::ArmarXObjects::NavigationMemory
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "Component.h"

#include <filesystem>
#include <fstream>
#include <iostream>
#include <iterator>

#include "ArmarXCore/core/Component.h"
#include "ArmarXCore/core/time/Metronome.h"
#include "ArmarXCore/core/time/forward_declarations.h"
#include <ArmarXCore/core/system/ArmarXDataPath.h>
#include <ArmarXCore/core/time/CycleUtil.h>
#include <ArmarXCore/libraries/DecoupledSingleComponent/Decoupled.h>

#include "RobotAPI/libraries/armem/core/MemoryID.h"
#include "RobotAPI/libraries/armem/core/wm/aron_conversions.h"
#include <RobotAPI/libraries/armem/client/query.h>
#include <RobotAPI/libraries/armem/core/aron_conversions.h>
#include <RobotAPI/libraries/armem/core/operations.h>
#include <RobotAPI/libraries/armem/server/ltm/disk/Memory.h>
#include <RobotAPI/libraries/armem/server/wm/memory_definitions.h>
#include <RobotAPI/libraries/aron/common/aron_conversions.h>

#include "Visu.h"
#include <armarx/navigation/algorithms/Costmap.h>
#include <armarx/navigation/algorithms/aron/Costmap.aron.generated.h>
#include <armarx/navigation/core/Graph.h>
#include <armarx/navigation/core/aron/Graph.aron.generated.h>
#include <armarx/navigation/core/aron/Location.aron.generated.h>
#include <armarx/navigation/core/aron/Trajectory.aron.generated.h>
#include <armarx/navigation/core/aron/Twist.aron.generated.h>
#include <armarx/navigation/human/aron/Human.aron.generated.h>
#include <armarx/navigation/graph/constants.h>
#include <armarx/navigation/location/constants.h>
#include <armarx/navigation/memory/constants.h>

namespace armarx::navigation::components::navigation_memory
{

    ARMARX_REGISTER_COMPONENT_EXECUTABLE(Component, Component::GetDefaultName());

    armarx::PropertyDefinitionsPtr
    Component::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr def =
            new ComponentPropertyDefinitions(getConfigIdentifier());

        // Publish to a topic (passing the TopicListenerPrx).
        // def->topic(myTopicListener);

        // Subscribe to a topic (passing the topic name).
        // def->topic<PlatformUnitListener>("MyTopic");

        // Use (and depend on) another component (passing the ComponentInterfacePrx).
        // def->component(myComponentProxy)


        workingMemory().name() = "Navigation";
        def->optional(properties.snapshotToLoad,
                      "p.snapshotToLoad",
                      "Memory snapshot to load at start up \n"
                      "(e.g. 'PriorKnowledgeData/navigation-graphs/snapshot').");

        def->optional(properties.locationGraph.visuLocations,
                      "p.locationGraph.visuLocation",
                      "Enable visualization of locations.");
        def->optional(properties.locationGraph.visuGraphEdges,
                      "p.locationGraph.visuGraphEdges",
                      "Enable visualization of navigation graph edges.");
        def->optional(properties.locationGraph.visuFrequency,
                      "p.locationGraph.visuFrequency",
                      "Visualization frequeny of locations and graph edges [Hz].");

        return def;
    }


    void
    Component::onInitComponent()
    {
        // Topics and properties defined above are automagically registered.

        // Keep debug observer data until calling `sendDebugObserverBatch()`.
        // (Requies the armarx::DebugObserverComponentPluginUser.)
        // setDebugObserverBatchModeEnabled(true);

        workingMemory().addCoreSegment(
            navigation::memory::constants::ParameterizationCoreSegmentName);

        workingMemory().addCoreSegment(navigation::memory::constants::CostmapCoreSegmentName,
                                       algorithms::arondto::Costmap::ToAronType());

        workingMemory().addCoreSegment("Results_GlobalPlanner",
                                       navigation::core::arondto::GlobalTrajectory::ToAronType());
        workingMemory().addCoreSegment("Results_LocalPlanner",
                                       navigation::core::arondto::LocalTrajectory::ToAronType());
        // workingMemory().addCoreSegment("Results_TrajectoryController",
        //                                navigation::core::arondto::Twist::ToAronType());
        // workingMemory().addCoreSegment("Results_SafetyController",
        //                                navigation::core::arondto::Twist::ToAronType());

        workingMemory().addCoreSegment("Events"); //, armem::example::ExampleData::ToAronType());
        // workingMemory.addCoreSegment("Exceptions");  //, armem::example::ExampleData::ToAronType());


        workingMemory().addCoreSegment(navigation::location::coreSegmentID.coreSegmentName,
                                       navigation::location::arondto::Location::ToAronType());
        workingMemory().addCoreSegment(navigation::graph::coreSegmentID.coreSegmentName,
                                       navigation::core::arondto::Graph::ToAronType());

        workingMemory().addCoreSegment(memory::constants::HumanCoreSegmentName,
                                       navigation::human::arondto::Human::ToAronType());

        // workingMemory().addCoreSegment(memory::constants::HumanGroupCoreSegmentName,
        //                                navigation::human::arondto::Human::ToAronType());

        if (not properties.snapshotToLoad.empty())
        {
            std::filesystem::path snapshotToLoadPath(
                ArmarXDataPath::resolvePath(properties.snapshotToLoad));
            if (std::filesystem::is_directory(snapshotToLoadPath))
            {
                // This section loads the snapshot specified by the scenario parameters
                // resolve the paths for the locations and graphs
                const std::filesystem::path graph = snapshotToLoadPath / "Graph";
                const std::filesystem::path locationsFilename =
                    snapshotToLoadPath / "locations.json";

                // remove date from folder name (if present)
                // Sometimes, we use the date before the snapshotname and use a symlink to the real data (e.g. R003 and 2022-03-01_R003)
                auto splitted = simox::alg::split(snapshotToLoadPath.filename().string(), "_");
                ARMARX_CHECK_GREATER(splitted.size(), 0);
                const std::string providerName = splitted[splitted.size() - 1];

                // This if statement loads the location. Each location is a single file (without extension). The filename is the name of the location.
                // The file contains json with the globalRobotPose (4x4 matrix) and relativeToObject information
                if (std::filesystem::is_regular_file(locationsFilename))
                {
                    armem::Commit c;
                    armem::MemoryID providerID = workingMemory().id();
                    providerID.coreSegmentName = "Location";
                    providerID.providerSegmentName = providerName;

                    const auto now = armem::Time::Now();

                    {
                        if (std::filesystem::is_regular_file(
                                locationsFilename)) // check if its a file (otherwise skip)
                        {
                            ARMARX_VERBOSE << "Loading " << locationsFilename;
                            std::ifstream ifs(locationsFilename);
                            const std::string content((std::istreambuf_iterator<char>(ifs)),
                                                      (std::istreambuf_iterator<char>()));

                            // parse location as json. All files in Location folder must be valid json objects!
                            nlohmann::json js = nlohmann::json::parse(content);

                            for (const auto& [locationMemoryIdStr, j] :
                                 js["locations"].get<std::map<std::string, nlohmann::json>>())
                            {
                                if (j.find("globalRobotPose") == j.end())
                                {
                                    ARMARX_WARNING << "The element '" << locationMemoryIdStr
                                                   << "' has no 'globalRobotPose' member. Skipping "
                                                      "this entity.";
                                    continue;
                                }

                                if (j.find("relativeToObject") == j.end())
                                {
                                    ARMARX_WARNING << "The element '" << locationMemoryIdStr
                                                   << "' has no 'relativeToObject' member. "
                                                      "Skipping this entity.";
                                    continue;
                                }

                                navigation::location::arondto::Location loc;
                                j.at("globalRobotPose")
                                    .get_to(loc.globalRobotPose); // load the 4x4 matrix

                                // TODO: All location I have seen were null.
                                // I don't know how this member should look like (von @Fabian Peller to @Fabian Reister)
                                loc.relativeToObject = std::nullopt;

                                // send commit to memory
                                auto& up = c.add();
                                up.confidence = 1.0;
                                up.timeCreated = now;
                                up.timeSent = now;
                                up.timeArrived = now;
                                up.entityID = armarx::armem::MemoryID(locationMemoryIdStr);
                                up.instancesData = {loc.toAron()};
                            }
                        }
                    }

                    // commit all locations at once
                    iceAdapter().commit(c);
                }

                const auto now = armem::Time::Now();

                // Next we load the graphs. The graph folder may contain several graphs, represented by different json files.
                // Each of those graphs contains a list of files representing the vertices. The filename is the vertice id (ideally starting at 0).
                // The file contains a json with the corresponding location name (as path) and the adjacent vertives (representing the directed outgoing edges).
                if (std::filesystem::is_directory(graph))
                {
                    armem::Commit c;
                    armem::MemoryID providerID = workingMemory().id();
                    providerID.coreSegmentName = memory::constants::GraphCoreSegmentName;
                    providerID.providerSegmentName = providerName;

                    for (const auto& graphdir : std::filesystem::directory_iterator(
                             graph)) // iterate over the different graphs (json files)
                    {
                        if (std::filesystem::is_regular_file(
                                graphdir.path())) // assure that its a json file
                        {
                            const std::string graphName = graphdir.path().stem();

                            navigation::core::arondto::Graph g;

                            ARMARX_VERBOSE << "Loading graph `" << graphName << " `from file `"
                                           << graphdir.path() << "`.";

                            std::ifstream ifs(graphdir.path());
                            const std::string content((std::istreambuf_iterator<char>(ifs)),
                                                      (std::istreambuf_iterator<char>()));

                            nlohmann::json j = nlohmann::json::parse(content);

                            const auto& jEdges = j.at("edges");
                            const auto& jVertices = j.at("vertices");

                            for (const auto& jVertex : jVertices)
                            {
                                const armarx::armem::MemoryID vertexLocationMemoryId(
                                    jVertex.at("locationID"));

                                armarx::navigation::core::arondto::Vertex v;
                                v.vertexID = jVertex.at("vertexID");
                                toAron(v.locationID, vertexLocationMemoryId);
                                toAron(v.locationID.timestamp, armem::Time::Invalid());
                                v.locationID.instanceIndex = 0;

                                g.vertices.push_back(v);
                            }

                            for (const auto& jEdge : jEdges)
                            {
                                std::pair<std::int64_t, std::int64_t> edgeSourceTargetPair;
                                jEdge.get_to(edgeSourceTargetPair);

                                armarx::navigation::core::arondto::Edge e;
                                e.sourceVertexID = edgeSourceTargetPair.first;
                                e.targetVertexID = edgeSourceTargetPair.second;
                                g.edges.push_back(e);
                            }

                            auto& up = c.add();
                            up.confidence = 1.0;
                            up.timeCreated = now;
                            up.timeSent = now;
                            up.timeArrived = now;
                            up.entityID = providerID.withEntityName(graphName);
                            up.instancesData = {g.toAron()};
                        }
                    }

                    // send graph to memory
                    iceAdapter().commit(c);
                }

                // LEGACY CODE (Not working anymore since the wm json export changed due to ltm updates and aron updates)
                // armem::wm::Memory memory = armem::server::ltm::disk::load(path.value());
                //armem::server::ltm::disk::Memory ltm(path.value());
                //armem::wm::Memory memory;
                //ltm.loadAll(memory);
                //workingMemory().update(armem::toCommit(memory));
                ARMARX_INFO << "Loaded NavigationMemory '" << properties.snapshotToLoad << "'";
            }
            else
            {
                ARMARX_WARNING << "Could not load NavigationMemory '" << properties.snapshotToLoad
                               << "'. Continue with empty memory.";
            }
        }
    }


    void
    Component::onConnectComponent()
    {
        // Do things after connecting to topics and components.


        /* (Requies the armarx::DebugObserverComponentPluginUser.)
        // Use the debug observer to log data over time.
        // The data can be viewed in the ObserverView and the LivePlotter.
        // (Before starting any threads, we don't need to lock mutexes.)
        {
            setDebugObserverDatafield("numBoxes", properties.numBoxes);
            setDebugObserverDatafield("boxLayerName", properties.boxLayerName);
            sendDebugObserverBatch();
        }
        */

        // Setup the remote GUI.
        {
            createRemoteGuiTab();
            RemoteGui_startRunningTask();
        }

        tasks.visuTask = new SimpleRunningTask<>([this]() { this->visuRun(); }, "Visualization");
        tasks.visuTask->start();
    }


    void
    Component::onDisconnectComponent()
    {
    }


    void
    Component::onExitComponent()
    {
    }


    std::string
    Component::getDefaultName() const
    {
        return GetDefaultName();
    }

    std::string
    Component::GetDefaultName()
    {
        return "navigation_memory";
    }


    void
    Component::createRemoteGuiTab()
    {
        using namespace armarx::RemoteGui::Client;

        // Setup the widgets.
        tab.locationGraph.setup(*this);


        // Setup the layout.
        VBoxLayout root = {tab.locationGraph.group, VSpacer()};
        RemoteGui_createTab(getName(), root, &tab);
    }


    void
    Component::RemoteGui_update()
    {
        tab.locationGraph.update(*this);
    }


    void
    Component::RemoteGuiTab::LocationGraph::setup(Component& owner)
    {
        using namespace armarx::RemoteGui::Client;
        GridLayout grid;
        int row = 0;
        {
            visuLocations.setValue(owner.properties.locationGraph.visuLocations);
            visuGraphEdges.setValue(owner.properties.locationGraph.visuLocations);

            grid.add(Label("Visualize Locations"), {row, 0}).add(visuLocations, {row, 1});
            ++row;

            grid.add(Label("Visualize Graph Edges"), {row, 0}).add(visuGraphEdges, {row, 1});
            ++row;
        }

        group = GroupBox({grid});
    }


    void
    Component::RemoteGuiTab::LocationGraph::update(Component& owner)
    {
        if (visuLocations.hasValueChanged() or visuGraphEdges.hasValueChanged())
        {
            std::scoped_lock lock(owner.propertiesMutex);
            owner.properties.locationGraph.visuLocations = visuLocations.getValue();
            owner.properties.locationGraph.visuGraphEdges = visuGraphEdges.getValue();
        }
    }


    void
    Component::visuRun()
    {
        memory::Visu visu{arviz,
                          workingMemory().getCoreSegment(navigation::location::coreSegmentID),
                          workingMemory().getCoreSegment(navigation::graph::coreSegmentID),
                          workingMemory().getCoreSegment(memory::constants::CostmapCoreSegmentName),
                          workingMemory().getCoreSegment(memory::constants::HumanCoreSegmentName)};


        Properties::LocationGraph p;
        {
            std::scoped_lock lock(propertiesMutex);
            p = properties.locationGraph;
        }

        Metronome metronome(Frequency::HertzDouble(p.visuFrequency));
        while (tasks.visuTask and not tasks.visuTask->isStopped())
        {
            ARMARX_TRACE;

            {
                std::scoped_lock lock(propertiesMutex);
                p = properties.locationGraph;
            }

            std::vector<viz::Layer> layers;

            // Locations
            visu.drawLocations(layers, p.visuLocations);

            // Graph Edges
            visu.drawGraphs(layers, p.visuGraphEdges);

            // Costmaps
            visu.drawCostmaps(layers, p.visuCostmaps);

            // Humans
            visu.drawHumans(layers, p.visuHumans, p.visuTransparent);

            arviz.commit(layers);

            metronome.waitForNextTick();
        }
    }

    bool
    store(const std::map<armem::MemoryID, core::Graph>& graphs,
          const std::filesystem::path& baseDirectory)
    {
        ARMARX_INFO << "Creating export directory `" << baseDirectory << "`.";
        std::filesystem::create_directories(baseDirectory);

        for (const auto& [memoryId, graph] : graphs)
        {
            const std::filesystem::path nestedSubDir =
                std::filesystem::path(memoryId.providerSegmentName) / memoryId.coreSegmentName;
            const std::filesystem::path dir = baseDirectory / nestedSubDir;

            std::filesystem::create_directories(dir);

            nlohmann::json j;

            // source -> target
            std::vector<std::pair<std::size_t, std::size_t>> outgoingEdges;
            std::transform(graph.m_edges.begin(),
                           graph.m_edges.end(),
                           std::back_inserter(outgoingEdges),
                           [](const auto& edge)
                           { return std::make_pair(edge.m_source, edge.m_target); });

            j["edges"] = outgoingEdges;
            j["vertices"] = {};

            for (const auto& vertex : graph.m_vertices)
            {

                armarx::armem::MemoryID locationId;
                fromAron(vertex.m_property.aron.locationID, locationId);

                nlohmann::json jVertex;
                jVertex["locationID"] = locationId.getEntityID().str();
                jVertex["vertexID"] = vertex.m_property.aron.vertexID;

                j["vertices"].push_back(jVertex);
            }

            // save to disk
            const std::filesystem::path filename = dir / (memoryId.entityName + ".json");
            ARMARX_VERBOSE << "Saving file `" << filename << "`.";
            std::ofstream ofs(filename);
            ofs << std::setw(4) << j;
        }

        return true;
    }

    bool
    store(const std::map<armem::MemoryID, location::arondto::Location>& locations,
          const std::filesystem::path& baseDirectory)
    {
        ARMARX_INFO << "Creating export directory `" << baseDirectory << "`.";
        std::filesystem::create_directories(baseDirectory);

        // key: provider id
        std::map<std::string, nlohmann::json> js;


        for (const auto& [memoryId, location] : locations)
        {
            auto& j = js[memoryId.providerSegmentName];

            if (j.count("locations") == 0)
            {
                j["locations"] = {};
            }

            nlohmann::json jLoc;

            jLoc["globalRobotPose"] = location.globalRobotPose;

            if (location.relativeToObject)
            {
                armarx::armem::MemoryID memoryId;
                fromAron(location.relativeToObject->objectInstanceID, memoryId);

                jLoc["relativeToObject"]["objectInstanceID"] = memoryId.str();
                jLoc["relativeToObject"]["relativeRobotPose"] =
                    location.relativeToObject->relativeRobotPose;
            }
            else
            {
                jLoc["relativeToObject"] = "null";
            }

            j["locations"][memoryId.getEntityID().str()] = jLoc;
        }

        // save to disk
        for (const auto& [providerId, j] : js)
        {
            const std::filesystem::path subDir = std::filesystem::path(providerId);
            const std::filesystem::path dir = baseDirectory / subDir;

            if (not std::filesystem::exists(dir))
            {
                std::filesystem::create_directories(dir);
            }

            const std::filesystem::path filename = dir / "locations.json";
            ARMARX_VERBOSE << "Saving file `" << filename << "`.";
            std::ofstream ofs(filename);
            ofs << std::setw(4) << j;
        }


        return true;
    }


    bool
    Component::storeLocationGraph(const armarx::data::PackagePath& packagePath,
                                         const Ice::Current& current)
    {
        armem::server::wm::CoreSegment& locationCoreSegment =
            workingMemory().getCoreSegment(navigation::location::coreSegmentID);
        armem::server::wm::CoreSegment& graphCoreSegment =
            workingMemory().getCoreSegment(navigation::graph::coreSegmentID);

        // obtain locations and graphs

        const std::map<armem::MemoryID, location::arondto::Location> locations =
            [&locationCoreSegment]()
        {
            std::map<armem::MemoryID, location::arondto::Location> locations;
            locationCoreSegment.doLocked(
                [&]()
                {
                    locationCoreSegment.forEachEntity(
                        [&](const armarx::armem::server::wm::Entity& entity) -> bool
                        {
                            if (const armarx::armem::server::wm::EntityInstance* instance =
                                    entity.findLatestInstance())
                            {
                                locations[entity.id()].fromAron(instance->data());
                            }

                            return true;
                        });

                    return true;
                });

            return locations;
        }();

        const auto graphs = [&graphCoreSegment]()
        {
            std::map<armem::MemoryID, core::Graph> graphs;
            graphCoreSegment.doLocked(
                [&]()
                {
                    graphCoreSegment.forEachEntity(
                        [&](const armarx::armem::server::wm::Entity& entity) -> bool
                        {
                            core::Graph& graph = graphs[entity.id()];
                            if (const armarx::armem::server::wm::EntityInstance* instance =
                                    entity.findLatestInstance())
                            {
                                navigation::core::arondto::Graph aron;
                                aron.fromAron(instance->data());
                                fromAron(aron, graph);
                            }

                            return true;
                        });
                });

            return graphs;
        }();


        // store on disk

        const std::filesystem::path baseDirectory = armarx::PackagePath(packagePath).toSystemPath();

        store(locations, baseDirectory);
        store(graphs, baseDirectory);

        return true;
    }


}  // namespace armarx::navigation::components::navigation_memory
