/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Timo Weberruß ( timo dot weberruss at student dot kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include <iostream>

#include <ArmarXCore/core/time/DateTime.h>
#include <ArmarXCore/core/time/Duration.h>
#include <ArmarXCore/interface/core/BasicVectorTypes.h>

#include <VisionX/libraries/armem_human/types.h>

#include <armarx/navigation/core/basic_types.h>
#include <armarx/navigation/human/HumanFilter.h>
#include <armarx/navigation/human/HumanTracker.h>
#include <armarx/navigation/human/types.h>

// test includes
#define BOOST_TEST_MODULE Navigation::ArmarXLibraries::human
#define ARMARX_BOOST_TEST

#include <armarx/navigation/Test.h>

using armarx::armem::human::HumanPose;
using armarx::armem::human::PoseKeypoint;
using armarx::navigation::core::Pose2D;
using CamMm = armarx::navigation::human::HumanTracker::CameraMeasurement;
using LaserMm = armarx::navigation::human::HumanTracker::LaserMeasurement;
using Eigen::Vector2f;
using Ellipsoid = armarx::armem::vision::Ellipsoid;
using Circle = armarx::armem::vision::Circle;

namespace armarx::navigation::human
{
    Eigen::Quaternionf
    quatFromEuler(float roll, float pitch, float yaw)
    {
        return Eigen::AngleAxisf(roll, Eigen::Vector3f::UnitX()) *
               Eigen::AngleAxisf(pitch, Eigen::Vector3f::UnitY()) *
               Eigen::AngleAxisf(yaw, Eigen::Vector3f::UnitZ());
    }

    /**
     * @brief generateCamMeasurements generates a series of linear camera measurements of a moving
     * human's head. It will start at startPosition and linearly move to endPosition over the course of
     * the given timespan, giving a vector with the given amount of steps inbetween.
     *
     * @param startPosition the 3D position of the human head at the first measurement
     * @param endPosition the 3D position of the human head at the last measurement
     * @param timespanMs the timespan in milliseconds over which the simulated human will move from start to end
     * @param steps how many measurements will be made (incl. start and end measurement)
     * @return a vector of <steps> measurements spread equally with regard to position and timespan
     */
    std::vector<CamMm>
    generateCamMeasurements(const Eigen::Vector3f& startPosition,
                            const Eigen::Vector3f& endPosition,
                            const Eigen::Quaternionf& orientation,
                            const int timespanMs,
                            const int steps)
    {
        const double timestepMs = static_cast<double>(timespanMs) / steps;
        const Eigen::Vector3f posDelta = (endPosition - startPosition) / steps;

        std::vector<CamMm> measurements = std::vector<CamMm>();
        for (int i = 0; i <= steps; i++)
        {
            const DateTime t = DateTime(Duration::MilliSeconds(i * timestepMs));

            const Eigen::Vector3f dummyPos = Eigen::Vector3f::Zero();
            const FramedPosition cameraDummyPos = FramedPosition(dummyPos, "", "");

            const Eigen::Vector3f newPos = startPosition + i * posDelta;
            const FramedPosition headPosition = FramedPosition(newPos, "", "");
            const FramedOrientation headOrientation = FramedOrientation(orientation, "", "");
            const PoseKeypoint head = PoseKeypoint{.label = "HEAD",
                                                   .confidence = 0.95,
                                                   .positionCamera = cameraDummyPos,
                                                   .positionGlobal = headPosition,
                                                   .orientationGlobal = headOrientation};
            const HumanPose pose = {.poseModelId = "posemodelid", .keypoints = {{"HEAD", head}}};
            const std::vector<armem::human::HumanPose> humanPoses = {pose};
            const CamMm camMm = {t, humanPoses};
            measurements.emplace_back(camMm);
        }
        return measurements;
    }

    LaserMm
    generateLaserMeasurement(const Eigen::Vector2f& pos, const DateTime& tLaser)
    {
        using Point = Eigen::Vector2f;

        // These are convex
        const std::vector<Point> relOffsets = {
            Point(0.01, 0.003), Point(-0.02, 0.007), Point(-0.01, -0.01), Point(0.015, 0.009)};

        const Eigen::Vector2f leftFootPos = pos + Eigen::Vector2f(0, 0.1);
        const Eigen::Vector2f rightFootPos = pos + Eigen::Vector2f(0, -0.1);

        std::vector<Point> leftFootPoints = std::vector<Point>();
        std::vector<Point> rightFootPoints = std::vector<Point>();
        int i = 0;
        for (const Point& offset : relOffsets)
        {
            leftFootPoints.emplace_back(leftFootPos + offset);
            rightFootPoints.emplace_back(rightFootPos + offset);
            i++;
        }
        std::vector<Point> leftFootHull = std::vector<Point>(leftFootPoints);
        std::vector<Point> rightFootHull = std::vector<Point>(rightFootPoints);

        // Add one more non-convex point to the cluster points
        leftFootPoints.emplace_back(leftFootPos + Point(0.002, 0.001));
        rightFootPoints.emplace_back(rightFootPos + Point(0.002, 0.003));

        using Pose = Eigen::Isometry3f;
        Pose leftFootPose = Pose::Identity();
        leftFootPose.translation() = Eigen::Vector3f(leftFootPos.x(), leftFootPos.y(), 0);
        //        leftFootPose.linear() = Eigen::Rotation2Df(orientation).toRotationMatrix();

        Pose rightFootPose = Pose::Identity();
        rightFootPose.translation() = Eigen::Vector3f(rightFootPos.x(), rightFootPos.y(), 0);
        ;
        //        rightFootPose.linear() = Eigen::Rotation2Df(orientation).toRotationMatrix();

        const Ellipsoid leftFootEllipse = {.pose = leftFootPose, .radii = Point(0.05, 0.04)};
        const Ellipsoid rightFootEllipse = {.pose = rightFootPose, .radii = Point(0.05, 0.04)};

        const Circle leftFootCircle = {.center = leftFootPos, .radius = 0.1};
        const Circle rightFootCircle = {.center = rightFootPos, .radius = 0.1};

        const Cluster leftFootCluster = {
            .convexHull = leftFootHull,
            .circle = leftFootCircle,
            .ellipsoid = leftFootEllipse,
            .chain = leftFootPoints, //don't use me, probably initialized wrong
            .points = leftFootPoints};
        const Cluster rightFootCluster = {
            .convexHull = rightFootHull,
            .circle = rightFootCircle,
            .ellipsoid = rightFootEllipse,
            .chain = rightFootPoints, //don't use me, probably initialized wrong
            .points = rightFootPoints};

        const std::vector<Cluster> clusters =
            std::vector<Cluster>{leftFootCluster, rightFootCluster};

        const LaserMm laserMm = {.detectionTime = tLaser, .clusters = clusters};
        return laserMm;
    }

    BOOST_AUTO_TEST_CASE(testLaserTrackingTwoFeetCluster)
    {
        HumanTracker tracker = HumanTracker();

        // PARAMETERS

        const Eigen::Vector3f initialPosition = Eigen::Vector3f(0, 0, 2) * 1000;
        const float orientation = M_PI_2;
        const Eigen::Quaternionf orientationQuat = quatFromEuler(0, 0, orientation);

        const std::int64_t timestepMs = 100;
        const Eigen::Vector3f movementSpeedMmPerSec = Eigen::Vector3f(1, 0, 0) * 1000;
        const int cameraSteps = 100;

        // CONVERSIONS
        const float timespanMs = timestepMs * cameraSteps;
        const Eigen::Vector3f endPosition =
            initialPosition + movementSpeedMmPerSec * timespanMs / 1000;
        const Eigen::Vector3f posDelta = (endPosition - initialPosition) / cameraSteps;

        // CAMERA MEASUREMENTS

        const std::vector<CamMm> camMeasurements = generateCamMeasurements(
            initialPosition, endPosition, orientationQuat, timespanMs, cameraSteps);

        for (const CamMm& measurement : camMeasurements)
        {
            tracker.update(measurement);

            int i = 0;
            for (const auto& trackedHuman : tracker.getTrackedHumans())
            {
                const Eigen::Vector2f tPos = trackedHuman.pose.translation();
                BOOST_TEST_MESSAGE("tracker now thinks human "
                                   << i << " is at x=(" << tPos.x() << "," << tPos.y()
                                   << ") moving with v=(" << trackedHuman.linearVelocity.x() << ","
                                   << trackedHuman.linearVelocity.y() << ")");
                i++;
            }
        }

        // LASER MEASUREMENT

        const DateTime tLaser = DateTime(Duration::MilliSeconds(timespanMs + timestepMs));
        Eigen::Vector2f laserPos = (endPosition + posDelta).segment(0, 2);
        const LaserMm laserMm = generateLaserMeasurement(laserPos, tLaser);

        tracker.update(laserMm);

        const std::vector<Human> humans = tracker.getTrackedHumans();
        BOOST_CHECK_EQUAL(humans.size(), 1); // should have detected the one human through laser
        const Human h = humans.at(0);
        BOOST_CHECK(h.pose.translation().x() - laserPos.x() < 0.05 * 1000);
        BOOST_CHECK(h.pose.translation().y() - laserPos.y() < 0.05 * 1000);
        BOOST_TEST_MESSAGE("final position: x=(" << h.pose.translation().x() << ","
                                                 << h.pose.translation().y() << ")");
        BOOST_TEST_MESSAGE("final velocity: v=(" << h.linearVelocity.x() << ","
                                                 << h.linearVelocity.y() << ")");
        BOOST_CHECK(h.linearVelocity.x() - movementSpeedMmPerSec.x() < 0.05 * 1000);
        BOOST_CHECK(h.linearVelocity.y() - movementSpeedMmPerSec.y() < 0.05 * 1000);
    }


    BOOST_AUTO_TEST_CASE(testRemoveOldHumans)
    {
        HumanTracker tracker = HumanTracker();
        auto zero = Eigen::Vector3f(0.f, 0.f, 0.f);
        auto measurements =
            generateCamMeasurements(zero, zero, Eigen::Quaternionf::Identity(), 0, 1);
        tracker.update(measurements.at(0));

        BOOST_CHECK_EQUAL(tracker.getTrackedHumans().size(), 1);

        CamMm late = {.detectionTime = DateTime(Duration::MilliSeconds(1000))};
        tracker.update(late);

        BOOST_CHECK_EQUAL(tracker.getTrackedHumans().size(), 0);
    }
} // namespace armarx::navigation::human
