/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Tobias Gröger ( tobias dot groeger at student dot kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#define BOOST_TEST_MODULE Navigation::ArmarXLibraries::human
#define ARMARX_BOOST_TEST

#include <cstdlib> /* srand, rand */
#include <ctime> /* time */

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/util/time.h>

#include <RobotAPI/libraries/ukfm/UnscentedKalmanFilter.h>

#include <armarx/navigation/Test.h>
#include <armarx/navigation/human/HumanSystemModel.h>
#include <manif/SO2.h>
#include <sciplot/sciplot.hpp>

namespace armarx::navigation::human
{
    using T = double; //TODO double or float?
    using Vector = Eigen::Matrix<T, 2, 1>;
    using SystemModelT = kalman_filter::SystemModelSE2<T>;

    constexpr T RAD = 10;
    constexpr long num_timesteps = 3000;
    constexpr T max_time = 1;
    constexpr T dt = max_time / num_timesteps;
    constexpr T c = (1 / max_time) * 2 * M_PI; // rotation per time

    constexpr T rot_noise_std = 0.1;
    constexpr T pos_noise_std = 1;
    constexpr T obs_rot_noise_std = 0.1;
    constexpr T obs_pos_noise_std = 1;
    constexpr T initial_offset_angle = 0.1 * 10 * M_PI / 180;
    const Vector initial_offet_pos = 0.1 * RAD * Vector(1, 0.5);


    void
    simulate_trajectory(std::vector<SystemModelT::StateT>& states,
                        std::vector<SystemModelT::ControlT>& omegas)
    {
        constexpr T startAngle = 0;

        SystemModelT::StateT state;
        state.pose = manif::SE2<T>(RAD, 0, startAngle);
        states.push_back(state);

        for (int i = 1; i < num_timesteps; i++)
        {
            const T t = dt * i;
            const T angle = t * c;

            const Vector last_pos = states.at(i - 1).pose.translation();
            //const Vector pos(angle, std::cos(angle));
            const Vector pos = RAD * Vector(std::cos(angle), std::sin(angle));

            BOOST_TEST(((last_pos - pos).norm() < 0.1 * RAD),
                       "position differs too much from last step: " << (last_pos - pos).norm());

            Eigen::Vector2d euclidean_velocity = RAD * Eigen::Vector2d{0, c};
            SystemModelT::ControlT control;
            //control.velocity.coeffs() << (pos - last_pos) / dt, c;
            control.velocity.coeffs() << euclidean_velocity, c;

            const SystemModelT::StateT propagated = SystemModelT::propagationFunction(
                state, control, SystemModelT::ControlNoiseT::Zero(), dt);

            state.pose = manif::SE2<T>(pos.x(), pos.y(), angle + startAngle);

            T pos_diff = (propagated.pose.translation() - state.pose.translation()).norm();
            BOOST_TEST(pos_diff < 2e-10,
                       "propagated position differs too much from real position: " << pos_diff);

            T rot_diff = propagated.pose.angle() - state.pose.angle();
            BOOST_TEST(rot_diff < 2e-10,
                       "propagated rotation differs too much from real orientation: " << rot_diff);

            states.push_back(propagated);


            // add noise
            control.velocity.coeffs().segment(0, 2) += pos_noise_std * Vector::Random();
            control.velocity.coeffs().segment(2, 1) +=
                rot_noise_std * Eigen::Matrix<T, 1, 1>::Random();
            omegas.push_back(control);
        }
    }

    void
    simulate_observation(const std::vector<SystemModelT::StateT>& states,
                         std::vector<SystemModelT::ObsT>& observations)
    {
        for (const auto& state : states)
        {
            SystemModelT::ObsT obs = SystemModelT::observationFunction(state);

            SystemModelT::ObsT noise = SystemModelT::ObsT::Random();
            noise.segment(0, 2) *= obs_pos_noise_std;
            noise.segment(2, 1) *= obs_rot_noise_std;

            BOOST_TEST_MESSAGE("SALT TEST: noise " << noise);

            observations.push_back(obs + noise);

            BOOST_TEST_MESSAGE("SALT TEST: state:\n"
                               << state.pose.x() << "\n"
                               << state.pose.y() << "\n"
                               << state.pose.angle());

            BOOST_TEST_MESSAGE("SALT TEST: state.log:\n" << state.pose.log().coeffs());

            BOOST_TEST_MESSAGE("SALT TEST: noisy obs:\n" << (obs + noise));

            SystemModelT::ObsT diff =
                (obs + noise) -
                SystemModelT::ObsT(state.pose.x(), state.pose.y(), state.pose.angle());

            BOOST_TEST_MESSAGE("\nSALT TEST: diff:\n" << diff);
        }
    }


    BOOST_AUTO_TEST_CASE(se2KalmanFilterTest)
    {
        srand(time(NULL));

        std::vector<SystemModelT::StateT> states;
        std::vector<SystemModelT::ControlT> omegas;
        std::vector<SystemModelT::ObsT> observations;
        simulate_trajectory(states, omegas);
        simulate_observation(states, observations);

        ARMARX_INFO << "Num States: " << states.size() << " Num Controls: " << omegas.size()
                    << " Num Obs: " << observations.size();

        UnscentedKalmanFilter<SystemModelT>::PropCovT Q =
            UnscentedKalmanFilter<SystemModelT>::PropCovT::Identity();
        Q.block<2, 2>(0, 0) *= pos_noise_std * pos_noise_std;
        Q.block<1, 1>(2, 2) *= rot_noise_std * rot_noise_std;

        UnscentedKalmanFilter<SystemModelT>::ObsCovT R =
            UnscentedKalmanFilter<SystemModelT>::ObsCovT::Identity();
        R.block<2, 2>(0, 0) *= obs_pos_noise_std * obs_pos_noise_std;
        R.block<1, 1>(2, 2) *= obs_rot_noise_std * obs_rot_noise_std;

        UnscentedKalmanFilter<SystemModelT>::StateCovT P0 =
            UnscentedKalmanFilter<SystemModelT>::StateCovT::Identity();
        P0.block<2, 2>(0, 0) *= initial_offet_pos.norm() * initial_offet_pos.norm();
        P0.block<1, 1>(2, 2) *= initial_offset_angle * initial_offset_angle;

        UnscentedKalmanFilter<SystemModelT>::AlphaT alpha;
        alpha(0) = 0.5; //0.8; // state
        alpha(1) = 0.5; //0.1; // control
        alpha(2) = 0.5; //0.1; // update

        SystemModelT::StateT state0;
        const Vector startPos = states.at(0).pose.translation() + initial_offet_pos;
        state0.pose = manif::SE2<T>{
            startPos.x(), startPos.y(), states.at(0).pose.angle() + initial_offset_angle};

        UnscentedKalmanFilter<SystemModelT> ukf(Q, R, alpha, state0, P0);

        std::vector<SystemModelT::StateT> ukf_states;
        std::vector<UnscentedKalmanFilter<SystemModelT>::StateCovT> ukf_Ps;
        ukf_states.push_back(state0);
        ukf_Ps.push_back(P0);

        std::vector<T> x_true, y_true, x_obs, y_obs, x_ukf, y_ukf;
        std::vector<T> a_true, a_obs, a_ukf;
        std::vector<T> x_ukf_full, y_ukf_full, a_ukf_full;
        std::vector<T> ukf_pos_var, ukf_rot_var, ukf_var;

        for (int i = 1; i < num_timesteps; i++)
        {
            // propagate
            TIMING_START(LOOP);
            TIMING_START(PROPAGATION);
            ukf.propagation(omegas.at(i - 1), dt);
            TIMING_END(PROPAGATION);
            if ((i - 1) % 1 == 0)
            {
                TIMING_START(UPDATE);
                ukf.update(observations.at(i));
                TIMING_END(UPDATE);
                TIMING_START(REST);

                BOOST_TEST_MESSAGE("Max Cov "
                                   << ukf.getCurrentStateCovariance().cwiseAbs().maxCoeff());
                BOOST_TEST_MESSAGE("Diag: " << (ukf.getCurrentStateCovariance() -
                                                Eigen::Matrix<T, 3, 3>::Identity())
                                                   .norm());

                manif::SE2Tangent<T> obs{observations.at(i)};
                x_obs.push_back(obs.exp().x());
                y_obs.push_back(obs.exp().y());
                a_obs.push_back(obs.exp().angle());

                x_ukf.push_back(ukf.getCurrentState().pose.x());
                y_ukf.push_back(ukf.getCurrentState().pose.y());
                a_ukf.push_back(ukf.getCurrentState().pose.angle());

                TIMING_END(REST);
            }
            x_true.push_back(states.at(i).pose.x());
            y_true.push_back(states.at(i).pose.y());
            a_true.push_back(states.at(i).pose.angle());

            x_ukf_full.push_back(ukf.getCurrentState().pose.x());
            y_ukf_full.push_back(ukf.getCurrentState().pose.y());
            a_ukf_full.push_back(ukf.getCurrentState().pose.angle());

            ukf_states.push_back(ukf.getCurrentState());
            ukf_Ps.push_back(ukf.getCurrentStateCovariance());

            ukf_pos_var.push_back(
                ukf.getCurrentStateCovariance().block<2, 2>(0, 0).cwiseAbs().maxCoeff());
            ukf_rot_var.push_back(
                ukf.getCurrentStateCovariance().block<1, 1>(2, 2).cwiseAbs().maxCoeff());
            ukf_var.push_back(ukf.getCurrentStateCovariance().cwiseAbs().maxCoeff());

            TIMING_END(LOOP);
        }

        sciplot::Plot2D pos_plot;
        pos_plot.xlabel("x");
        pos_plot.ylabel("y");
        pos_plot.legend().atOutsideBottom().displayHorizontal().displayExpandWidthBy(2);
        pos_plot.xrange(-RAD * 1.2, RAD * 1.2).yrange(-RAD * 1.2, RAD * 1.2);
        //pos_plot.xrange(-4, 4).yrange(-4, 4);

        pos_plot.drawCurve(x_true, y_true).label("True");
        pos_plot.drawCurve(x_obs, y_obs).label("Obs");
        pos_plot.drawCurveWithPoints(x_ukf, y_ukf).label("UKF");
        pos_plot.drawCurve(x_ukf_full, y_ukf_full).label("Full UKF");


        sciplot::Plot2D orientation_plot;
        orientation_plot.xlabel("t");
        orientation_plot.ylabel("rot");
        orientation_plot.legend().atOutsideBottom().displayHorizontal().displayExpandWidthBy(2);
        orientation_plot.xrange(-0.1, 1.1).yrange(-4, 4);

        sciplot::Vec x = sciplot::linspace(0.0, 1.0, a_obs.size());
        sciplot::Vec x_t = sciplot::linspace(0.0, 1.0, a_true.size());
        orientation_plot.drawCurve(x_t, a_true).label("True").lineWidth(1);
        orientation_plot.drawCurve(x, a_obs).label("Obs").lineWidth(1);
        orientation_plot.drawCurve(x, a_ukf).label("UKF").lineWidth(1);
        orientation_plot.drawCurve(x_t, a_ukf_full).label("Full UKF").lineWidth(1);


        sciplot::Plot2D error_plot;
        error_plot.xlabel("t");
        error_plot.ylabel("err");
        error_plot.legend().atOutsideBottom().displayHorizontal().displayExpandWidthBy(2);
        error_plot.xrange(-0.1, 1.1).yrange("*", "*");

        error_plot.drawCurve(x_t, ukf_pos_var).label("Pos variance").lineWidth(1);
        error_plot.drawCurve(x_t, ukf_rot_var).label("Rot variance").lineWidth(1);
        error_plot.drawCurve(x_t, ukf_var).label("Total variance").lineWidth(1);


        // Create figure to hold plot
        sciplot::Figure fig = {{pos_plot, orientation_plot}, {error_plot}};
        // Create canvas to hold figure
        sciplot::Canvas canvas = {{fig}};
        canvas.size(1000, 1000);

        // Show the plot in a pop-up window
        canvas.show();

        // Save the plot to a PDF file
        canvas.save("se2_kalman_output.pdf");
    }

} // namespace armarx::navigation::human
